/*
 * Strawberry Music Player
 * Copyright 2010, David Sansome <me@davidsansome.com>
 * Copyright 2018-2021, Jonas Kvinge <jonas@jkvinge.net>
 *
 * Strawberry is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Strawberry is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Strawberry.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef PLATFORMINTERFACE_H
#define PLATFORMINTERFACE_H

#include <QObject>
#include <QString>

class PlatformInterface {
 public:
  PlatformInterface() = default;
  virtual ~PlatformInterface() {}

  // Called when the application should show itself.
  virtual void Activate() = 0;
  virtual bool LoadUrl(const QString &url) = 0;

 private:
  Q_DISABLE_COPY(PlatformInterface)
};

#endif  // PLATFORMINTERFACE_H
