/*
* Strawberry Music Player
* Copyright 2021-2023, Jonas Kvinge <jonas@jkvinge.net>
*
* Strawberry is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* Strawberry is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with Strawberry.  If not, see <http://www.gnu.org/licenses/>.
*
*/

#ifndef ENGINEDEVICE_H
#define ENGINEDEVICE_H

#include <QList>
#include <QVariant>
#include <QString>

class EngineDevice {
 public:
  explicit EngineDevice();
  QString GuessIconName() const;
  QString description;
  QVariant value;
  QString iconname;
  int card;
  int device;
};
using EngineDeviceList = QList<EngineDevice>;

#endif  // ENGINEDEVICE_H
