ALTER TABLE songs ADD COLUMN acoustid_id TEXT;

ALTER TABLE songs ADD COLUMN acoustid_fingerprint TEXT;

ALTER TABLE songs ADD COLUMN musicbrainz_album_artist_id TEXT;

ALTER TABLE songs ADD COLUMN musicbrainz_artist_id TEXT;

ALTER TABLE songs ADD COLUMN musicbrainz_original_artist_id TEXT;

ALTER TABLE songs ADD COLUMN musicbrainz_album_id TEXT;

ALTER TABLE songs ADD COLUMN musicbrainz_original_album_id TEXT;

ALTER TABLE songs ADD COLUMN musicbrainz_recording_id TEXT;

ALTER TABLE songs ADD COLUMN musicbrainz_track_id TEXT;

ALTER TABLE songs ADD COLUMN musicbrainz_disc_id TEXT;

ALTER TABLE songs ADD COLUMN musicbrainz_release_group_id TEXT;

ALTER TABLE songs ADD COLUMN musicbrainz_work_id TEXT;

ALTER TABLE subsonic_songs ADD COLUMN acoustid_id TEXT;

ALTER TABLE subsonic_songs ADD COLUMN acoustid_fingerprint TEXT;

ALTER TABLE subsonic_songs ADD COLUMN musicbrainz_album_artist_id TEXT;

ALTER TABLE subsonic_songs ADD COLUMN musicbrainz_artist_id TEXT;

ALTER TABLE subsonic_songs ADD COLUMN musicbrainz_original_artist_id TEXT;

ALTER TABLE subsonic_songs ADD COLUMN musicbrainz_album_id TEXT;

ALTER TABLE subsonic_songs ADD COLUMN musicbrainz_original_album_id TEXT;

ALTER TABLE subsonic_songs ADD COLUMN musicbrainz_recording_id TEXT;

ALTER TABLE subsonic_songs ADD COLUMN musicbrainz_track_id TEXT;

ALTER TABLE subsonic_songs ADD COLUMN musicbrainz_disc_id TEXT;

ALTER TABLE subsonic_songs ADD COLUMN musicbrainz_release_group_id TEXT;

ALTER TABLE subsonic_songs ADD COLUMN musicbrainz_work_id TEXT;

ALTER TABLE tidal_artists_songs ADD COLUMN acoustid_id TEXT;

ALTER TABLE tidal_artists_songs ADD COLUMN acoustid_fingerprint TEXT;

ALTER TABLE tidal_artists_songs ADD COLUMN musicbrainz_album_artist_id TEXT;

ALTER TABLE tidal_artists_songs ADD COLUMN musicbrainz_artist_id TEXT;

ALTER TABLE tidal_artists_songs ADD COLUMN musicbrainz_original_artist_id TEXT;

ALTER TABLE tidal_artists_songs ADD COLUMN musicbrainz_album_id TEXT;

ALTER TABLE tidal_artists_songs ADD COLUMN musicbrainz_original_album_id TEXT;

ALTER TABLE tidal_artists_songs ADD COLUMN musicbrainz_recording_id TEXT;

ALTER TABLE tidal_artists_songs ADD COLUMN musicbrainz_track_id TEXT;

ALTER TABLE tidal_artists_songs ADD COLUMN musicbrainz_disc_id TEXT;

ALTER TABLE tidal_artists_songs ADD COLUMN musicbrainz_release_group_id TEXT;

ALTER TABLE tidal_artists_songs ADD COLUMN musicbrainz_work_id TEXT;

ALTER TABLE tidal_albums_songs ADD COLUMN acoustid_id TEXT;

ALTER TABLE tidal_albums_songs ADD COLUMN acoustid_fingerprint TEXT;

ALTER TABLE tidal_albums_songs ADD COLUMN musicbrainz_album_artist_id TEXT;

ALTER TABLE tidal_albums_songs ADD COLUMN musicbrainz_artist_id TEXT;

ALTER TABLE tidal_albums_songs ADD COLUMN musicbrainz_original_artist_id TEXT;

ALTER TABLE tidal_albums_songs ADD COLUMN musicbrainz_album_id TEXT;

ALTER TABLE tidal_albums_songs ADD COLUMN musicbrainz_original_album_id TEXT;

ALTER TABLE tidal_albums_songs ADD COLUMN musicbrainz_recording_id TEXT;

ALTER TABLE tidal_albums_songs ADD COLUMN musicbrainz_track_id TEXT;

ALTER TABLE tidal_albums_songs ADD COLUMN musicbrainz_disc_id TEXT;

ALTER TABLE tidal_albums_songs ADD COLUMN musicbrainz_release_group_id TEXT;

ALTER TABLE tidal_albums_songs ADD COLUMN musicbrainz_work_id TEXT;

ALTER TABLE tidal_songs ADD COLUMN acoustid_id TEXT;

ALTER TABLE tidal_songs ADD COLUMN acoustid_fingerprint TEXT;

ALTER TABLE tidal_songs ADD COLUMN musicbrainz_album_artist_id TEXT;

ALTER TABLE tidal_songs ADD COLUMN musicbrainz_artist_id TEXT;

ALTER TABLE tidal_songs ADD COLUMN musicbrainz_original_artist_id TEXT;

ALTER TABLE tidal_songs ADD COLUMN musicbrainz_album_id TEXT;

ALTER TABLE tidal_songs ADD COLUMN musicbrainz_original_album_id TEXT;

ALTER TABLE tidal_songs ADD COLUMN musicbrainz_recording_id TEXT;

ALTER TABLE tidal_songs ADD COLUMN musicbrainz_track_id TEXT;

ALTER TABLE tidal_songs ADD COLUMN musicbrainz_disc_id TEXT;

ALTER TABLE tidal_songs ADD COLUMN musicbrainz_release_group_id TEXT;

ALTER TABLE tidal_songs ADD COLUMN musicbrainz_work_id TEXT;

ALTER TABLE qobuz_artists_songs ADD COLUMN acoustid_id TEXT;

ALTER TABLE qobuz_artists_songs ADD COLUMN acoustid_fingerprint TEXT;

ALTER TABLE qobuz_artists_songs ADD COLUMN musicbrainz_album_artist_id TEXT;

ALTER TABLE qobuz_artists_songs ADD COLUMN musicbrainz_artist_id TEXT;

ALTER TABLE qobuz_artists_songs ADD COLUMN musicbrainz_original_artist_id TEXT;

ALTER TABLE qobuz_artists_songs ADD COLUMN musicbrainz_album_id TEXT;

ALTER TABLE qobuz_artists_songs ADD COLUMN musicbrainz_original_album_id TEXT;

ALTER TABLE qobuz_artists_songs ADD COLUMN musicbrainz_recording_id TEXT;

ALTER TABLE qobuz_artists_songs ADD COLUMN musicbrainz_track_id TEXT;

ALTER TABLE qobuz_artists_songs ADD COLUMN musicbrainz_disc_id TEXT;

ALTER TABLE qobuz_artists_songs ADD COLUMN musicbrainz_release_group_id TEXT;

ALTER TABLE qobuz_artists_songs ADD COLUMN musicbrainz_work_id TEXT;

ALTER TABLE qobuz_albums_songs ADD COLUMN acoustid_id TEXT;

ALTER TABLE qobuz_albums_songs ADD COLUMN acoustid_fingerprint TEXT;

ALTER TABLE qobuz_albums_songs ADD COLUMN musicbrainz_album_artist_id TEXT;

ALTER TABLE qobuz_albums_songs ADD COLUMN musicbrainz_artist_id TEXT;

ALTER TABLE qobuz_albums_songs ADD COLUMN musicbrainz_original_artist_id TEXT;

ALTER TABLE qobuz_albums_songs ADD COLUMN musicbrainz_album_id TEXT;

ALTER TABLE qobuz_albums_songs ADD COLUMN musicbrainz_original_album_id TEXT;

ALTER TABLE qobuz_albums_songs ADD COLUMN musicbrainz_recording_id TEXT;

ALTER TABLE qobuz_albums_songs ADD COLUMN musicbrainz_track_id TEXT;

ALTER TABLE qobuz_albums_songs ADD COLUMN musicbrainz_disc_id TEXT;

ALTER TABLE qobuz_albums_songs ADD COLUMN musicbrainz_release_group_id TEXT;

ALTER TABLE qobuz_albums_songs ADD COLUMN musicbrainz_work_id TEXT;

ALTER TABLE qobuz_songs ADD COLUMN acoustid_id TEXT;

ALTER TABLE qobuz_songs ADD COLUMN acoustid_fingerprint TEXT;

ALTER TABLE qobuz_songs ADD COLUMN musicbrainz_album_artist_id TEXT;

ALTER TABLE qobuz_songs ADD COLUMN musicbrainz_artist_id TEXT;

ALTER TABLE qobuz_songs ADD COLUMN musicbrainz_original_artist_id TEXT;

ALTER TABLE qobuz_songs ADD COLUMN musicbrainz_album_id TEXT;

ALTER TABLE qobuz_songs ADD COLUMN musicbrainz_original_album_id TEXT;

ALTER TABLE qobuz_songs ADD COLUMN musicbrainz_recording_id TEXT;

ALTER TABLE qobuz_songs ADD COLUMN musicbrainz_track_id TEXT;

ALTER TABLE qobuz_songs ADD COLUMN musicbrainz_disc_id TEXT;

ALTER TABLE qobuz_songs ADD COLUMN musicbrainz_release_group_id TEXT;

ALTER TABLE qobuz_songs ADD COLUMN musicbrainz_work_id TEXT;

ALTER TABLE playlist_items ADD COLUMN acoustid_id TEXT;

ALTER TABLE playlist_items ADD COLUMN acoustid_fingerprint TEXT;

ALTER TABLE playlist_items ADD COLUMN musicbrainz_album_artist_id TEXT;

ALTER TABLE playlist_items ADD COLUMN musicbrainz_artist_id TEXT;

ALTER TABLE playlist_items ADD COLUMN musicbrainz_original_artist_id TEXT;

ALTER TABLE playlist_items ADD COLUMN musicbrainz_album_id TEXT;

ALTER TABLE playlist_items ADD COLUMN musicbrainz_original_album_id TEXT;

ALTER TABLE playlist_items ADD COLUMN musicbrainz_recording_id TEXT;

ALTER TABLE playlist_items ADD COLUMN musicbrainz_track_id TEXT;

ALTER TABLE playlist_items ADD COLUMN musicbrainz_disc_id TEXT;

ALTER TABLE playlist_items ADD COLUMN musicbrainz_release_group_id TEXT;

ALTER TABLE playlist_items ADD COLUMN musicbrainz_work_id TEXT;

UPDATE schema_version SET version=16;
