/*
 * Strawberry Music Player
 * This file was part of Clementine.
 * Copyright 2010, David Sansome <me@davidsansome.com>
 *
 * Strawberry is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Strawberry is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Strawberry.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef COVERSEARCHSTATISTICS_H
#define COVERSEARCHSTATISTICS_H

#include "config.h"

#include <QtGlobal>
#include <QList>
#include <QMap>
#include <QString>

struct CoverSearchStatistics {

  explicit CoverSearchStatistics();

  CoverSearchStatistics &operator+=(const CoverSearchStatistics &other);

  quint64 network_requests_made_;
  quint64 bytes_transferred_;
  QMap<QString, quint64> total_images_by_provider_;
  QMap<QString, quint64> chosen_images_by_provider_;

  quint64 chosen_images_;
  quint64 missing_images_;

  quint64 chosen_width_;
  quint64 chosen_height_;

  QString AverageDimensions() const;
};

#endif  // COVERSEARCHSTATISTICS_H
