#frame {
  border: 1px solid palette(mid);
  border-radius: 5px;
}

#container {
  margin: 2px;
}

#remove {
  border-top-left-radius: 0px;
  border-top-right-radius: 5px;
  border-bottom-left-radius: 0px;
  border-bottom-right-radius: 5px;
  border: 0px solid transparent;
  background: qlineargradient(x1:0, y1:0, x2:0, y2:1,
    stop:0   %light,
    stop:0.4 %light,
    stop:0.6 %dark,
    stop:1   %dark);

  margin-left: 5px;
  padding: 0px 5px;
}

#remove:hover {
  background: qlineargradient(x1:0, y1:0, x2:0, y2:1,
    stop:0   %light2,
    stop:0.4 %light2,
    stop:0.6 %base,
    stop:1   %base);
  border: 0px solid transparent;
}

#remove:pressed {
  background: qlineargradient(x1:0, y1:0, x2:0, y2:1,
    stop:0   %base,
    stop:0.4 %base,
    stop:0.6 %light2,
    stop:1   %light2);
  border: 0px solid transparent;
}
